unit SForm;

interface

uses Windows, Messages, CommDlg, CommCtrl;

type
  TNotifyEvent = procedure (Sender: TObject) of object;

  TWinControl = class(TObject)
  private
    FHandle: HWnd;
    function GetClientRect: TRect;
    procedure SetClientRect(Value: TRect);
    function GetText: string;
    procedure SetText(const Value: string);
    function GetBoundsRect: TRect; virtual;
    function GetEnabled: Boolean;
    procedure SetEnabled(Value: Boolean);
    function GetVisible: Boolean;
  public
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
    procedure SetSize(AWidth, AHeight: Integer);
    procedure SetPosition(ALeft, ATop: Integer);
    procedure Show;
    procedure Hide;
    procedure Update;
    procedure Invalidate;
    function Perform(Msg: Cardinal; WParam, LParam: Longint): Longint;
    function ScreenToClient(const Point: TPoint): TPoint;
    property ClientRect: TRect read GetClientRect write SetClientRect;
    property BoundsRect: TRect read GetBoundsRect;
    property Handle: HWnd read FHandle write FHandle;
    property Text: string read GetText write SetText;
    property Enabled: Boolean read GetEnabled write SetEnabled;
    property Visible: Boolean read GetVisible;
  end;

  TCommonControl = class(TWinControl)
    OrgWndProc: TFNWndProc;
    procedure CreateWnd(Parent: HWND; ALeft, ATop, AWidth, AHeight: Integer); virtual;
    procedure DefaultHandler(var Message); override;
  end;

  TSForm = class(TWinControl)
  private
    FSaveWP: TWindowPlacement;
    FSaveWS: Longint;
    FSaveStayOnTop: Boolean;
    FRegisteredClass: Boolean;
    FStayOnTop: Boolean;
    FFullScreen: Boolean;
    FBorder: Boolean;
    procedure SetStayOnTop(Value: Boolean);
    procedure SetFullScreen(Value: Boolean);
    procedure SetBorder(Value: Boolean);
    function GetMaximized: Boolean;
    function GetMinimized: Boolean;
  public
    Style: DWORD;
    ExStyle: DWORD;
    Brush: HBRUSH;
    Icon: HICON;
    Cursor: HCURSOR;
    ClassStyle: UINT;
    constructor Create(const Caption: string; ALeft, ATop, AWidth, AHeight: Integer); virtual;
    constructor CreateDefaultPos(const Caption: string; AWidth, AHeight: Integer);
    destructor Destroy; override;
    procedure Show;
    procedure Close;
    procedure Restore;
    procedure DefaultHandler(var Message); override;
    procedure InitParams; virtual;
    procedure CreateWnd(const Caption: string; ALeft, ATop, AWidth, AHeight: Integer); virtual;
    property StayOnTop: Boolean read FStayOnTop write SetStayOnTop;
    property FullScreen: Boolean read FFullScreen write SetFullScreen;
    property Border: Boolean read FBorder write SetBorder;
    property Maximized: Boolean read GetMaximized;
    property Minimized: Boolean read GetMinimized;
  end;

  TSDlg = class(TWinControl)
  private
    FParent: HWND;
    FTemplateName: string;
    FItem: TWinControl;
    function GetItem(Id: Integer): TWinControl;
  public
    constructor Create;
    constructor CreateByResource(Parent: HWND; const TemplateName: string); virtual;
    destructor Destroy; override;
    property Items[Id: Integer]: TWinControl read GetItem;
  end;

  TSModalDlg = class(TSDlg)
  public
    function ShowModal: Boolean;
    procedure Close(ResultCode: Integer);
  end;

  TSModelessDlg = class(TSDlg)
  public
    procedure ShowModeless;
    procedure Close;
  end;

  TSPropSheet = class(TObject)
  private
    FCaption: string;
    FPages: PPropSheetPage;
    FHeader: TPropSheetHeader;
  public
    constructor Create(Parent: HWND; const Caption: string; Pages: array of TSDlg);
    destructor Destroy; override;
    function ShowModal: Boolean;
  end;

  TSOpenDlg = class(TObject)
  private
    FParent: HWND;
    FFilter: string;
    FFileName: string;
  public
    constructor Create(Parent: HWND; const Filter: array of string; const FileName: string);
    function ShowModal(var AFileName: string): Boolean;
  end;

  TSSaveDlg = class(TSOpenDlg)
  public
    function ShowModal(var AFileName: string): Boolean;
  end;

  TSMenu = class(TObject)
  private
    FHandle: HMENU;
    function GetFlag(const Caption: string; Id: Integer): Integer;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AppendItem(const Caption: string; Id: Integer);
    procedure InsertItem(Index: Integer; const Caption: string; Id: Integer);
    procedure GrayItem(Id: Integer);
    procedure GrayItemByPos(Index: Integer);
    procedure CheckItem(Id: Integer; Checked: Boolean);
    function IndexOfItem(Id: Integer): Integer;
    property Handle: HMENU read FHandle;
  end;

  TSPopupMenu = class(TSMenu)
  public
    constructor Create;
    procedure Popup(Parent: HWND; const pt: TPoint);
  end;

  TSTooltip = class(TObject)
  private
    FHandle: HWnd;
    FParent: HWnd;
    FTime: Integer;
  protected
    procedure SetTime(const Value: Integer);
    procedure SetText(const Value: string);
  public
    constructor Create(Parent: HWnd); virtual;
    destructor Destroy; override;
    procedure Show;
    procedure Hide;
    procedure UpdateRect;
    procedure SetRect(const Rect: TRect);
    property Handle: HWnd read FHandle;
    property Time: Integer read FTime write SetTime;
    property Text: string write SetText;
  end;

function GetContextMenuPos(Handle: HWND; LParam: LPARAM; SelPosX, SelPosY: Longint; var pt: TPoint): Boolean;

procedure SetMainForm(const Form: TSForm);

function MessageLoop: Integer;
function MessageLoopWithDialog(var hDlg: HWND): Integer;
function MessageLoopWithAccel(hWnd: HWND; var hAccel: HACCEL): Integer;
function MessageLoopWithAccelAndDialog(hWnd: HWND; hAccel: HACCEL; var hDlg: HWND): Integer;

function AlreadyRun(const AppName: string): Boolean;
procedure FormatMsgBox(hWnd: HWND; const Caption: string; const Text: string;
  const Args: array of string; Flags: Integer);
procedure SetWindowCenter(hWnd: HWND);
function OsVersionInfo: TOsVersionInfo;
procedure AboutBox(hWnd: HWND; const Caption: string; const Text: string);
function MsgBox(hWnd: HWND; const Caption: string; const Text: string): Integer;
procedure ShowMsg(const Msg: string);

function Rect(ALeft, ATop, ARight, ABottom: Integer): TRect;
function Point(AX, AY: Integer): TPoint;

var
  MsgMouseWheel: UINT; // for Win95

implementation

const
  PropName = 'OBJECT';


{ TWinControl }

function TWinControl.GetClientRect: TRect;
begin
  Windows.GetClientRect(FHandle, Result);
end;

procedure TWinControl.SetClientRect(Value: TRect);
begin
  AdjustWindowRectEx(Value, GetWindowLong(FHandle, GWL_STYLE),
    (GetParent(FHandle) = 0) and (GetMenu(FHandle) <> 0),
    GetWindowLong(FHandle, GWL_EXSTYLE));
  with Value do
    SetSize(Right-Left, Bottom-Top);
end;

procedure TWinControl.Setbounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  MoveWindow(FHandle, ALeft, ATop, AWidth, AHeight, True);
end;

procedure TWinControl.SetSize(AWidth, AHeight: Integer);
begin
  SetWindowPos(FHandle, 0, 0, 0, AWidth, AHeight, SWP_NOZORDER or SWP_NOMOVE or SWP_NOACTIVATE);
end;

procedure TWinControl.SetPosition(ALeft, ATop: Integer);
begin
  SetWindowPos(FHandle, 0, ALeft, ATop, 0, 0, SWP_NOZORDER or SWP_NOSIZE or SWP_NOACTIVATE);
end;

function TWinControl.GetText: string;
var
  len: Integer;
begin
  len := GetWindowTextLength(FHandle);
  SetString(Result, nil, len);
  if len <> 0 then
    GetWindowText(FHandle, PChar(Result), len+1); // +1null terminate
end;

procedure TWinControl.SetText(const Value: string);
begin
  SetWindowText(FHandle, PChar(Value));
end;

function TWinControl.GetBoundsRect: TRect;
begin
  GetWindowRect(FHandle, Result);
end;

procedure TWinControl.Show;
begin
  ShowWindow(FHandle, SW_SHOW);
end;

procedure TWinControl.Hide;
begin
  ShowWindow(FHandle, SW_HIDE);
end;

procedure TWinControl.Update;
begin
  UpdateWindow(FHandle);
end;

procedure TWinControl.Invalidate;
begin
  InvalidateRect(FHandle, nil, True);
end;

function TWinControl.Perform(Msg: Cardinal; WParam, LParam: Longint): Longint;
begin
  Result := SendMessage(FHandle, Msg, WParam, LParam);
end;

function TWinControl.GetEnabled: Boolean;
begin
  Result := IsWindowEnabled(FHandle);
end;

procedure TWinControl.SetEnabled(Value: Boolean);
begin
  EnableWindow(FHandle, Value);
end;

function TWinControl.ScreenToClient(const Point: TPoint): TPoint;
begin
  Result := Point;
  Windows.ScreenToClient(FHandle, Result);
end;

function TWinControl.GetVisible: Boolean;
begin
  Result := IsWindowVisible(FHandle);
end;

{ TCommonControl }

function ComCtlWindowProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  CommonControl: TCommonControl;
  Message: TMessage;
begin
  CommonControl := TCommonControl(GetProp(hWnd, PropName));
  if Assigned(CommonControl) then
  begin
    Message.Msg := Msg;
    Message.WParam := wParam;
    Message.LParam := lParam;
    Message.Result := 0;
    CommonControl.Dispatch(Message);
    Result := Message.Result;
  end else
    Result := 0;
end;

procedure TCommonControl.DefaultHandler(var Message);
begin
  with TMessage(Message) do
    case Msg of
      WM_NCDESTROY:
        begin
          RemoveProp(FHandle, PropName);
          SetWindowLong(FHandle, GWL_WNDPROC, Integer(@OrgWndProc));
          Result := 0;
        end;
    else
      Result := CallWindowProc(OrgWndProc, FHandle, Msg, wParam, lParam);
    end;
end;

procedure TCommonControl.CreateWnd(Parent: HWND; ALeft, ATop, AWidth, AHeight: Integer);
begin
  SetProp(FHandle, PropName, Integer(Self));
  OrgWndProc := TFnWndProc(SetWindowLong(FHandle, GWL_WNDPROC, Integer(@ComCtlWindowProc)));
end;


{ TSForm }

var
  MainForm: TSForm;

procedure SetMainForm(const Form: TSForm);
begin
  MainForm := Form;
end;

function SFormWindowProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  Form: TSForm;
  Message: TMessage;
begin
  case Msg of
    WM_NCCREATE:
      begin
        SetProp(hWnd, PropName, Integer(PCreateStruct(lParam).lpCreateParams));
        TSForm(PCreateStruct(lParam).lpCreateParams).Handle := hWnd;
        Result := DefWindowProc(hWnd, Msg, wParam, lParam);
      end;
    WM_NCDESTROY:
      begin
        TSForm(GetProp(hWnd, PropName)).Handle := 0;
        RemoveProp(hWnd, PropName);
        Result := DefWindowProc(hWnd, Msg, wParam, lParam);
      end;
  else
    Form := TSForm(GetProp(hWnd, PropName));
    if Assigned(Form) then
    begin
      Message.Msg := Msg;
      Message.WParam := wParam;
      Message.LParam := lParam;
      Message.Result := 0;
      Form.Dispatch(Message);
      Result := Message.Result;
      if (Msg = WM_DESTROY) and (Form = MainForm) then
        PostQuitMessage(0);
    end else
      Result := 0;
  end;
end;

constructor TSForm.Create(const Caption: string; ALeft, ATop, AWidth, AHeight: Integer);
begin
  InitParams;
  CreateWnd(Caption, ALeft, ATop, AWidth, AHeight);
end;

constructor TSForm.CreateDefaultPos(const Caption: string; AWidth, AHeight: Integer);
begin
  Create(Caption, Integer(CW_USEDEFAULT), 0, AWidth, AHeight);
end;

destructor TSForm.Destroy;
begin
  if FHandle <> 0 then
    Close;
  inherited;
end;

procedure TSForm.InitParams;
begin
  FBorder := True;
  Brush := COLOR_3DFACE+1;
  Icon := LoadIcon(HInstance, 'MAINICON');
  Cursor := LoadCursor(0, IDC_ARROW);
  Style := WS_OVERLAPPEDWINDOW;
  ClassStyle := CS_DBLCLKS;
end;

procedure TSForm.DefaultHandler(var Message);
begin
  with TMessage(Message) do
    Result := DefWindowProc(FHandle, Msg, WParam, LParam);
end;

procedure TSForm.CreateWnd(const Caption: string; ALeft, ATop, AWidth, AHeight: Integer);
var
  wc: TWndClass;
  AClassName: string;
begin
  AClassName := ClassName;
  if not FRegisteredClass then
  begin
    FillChar(wc, SizeOf(wc), 0);
    wc.style         := ClassStyle;
    wc.lpfnWndProc   := @SFormWindowProc;
    wc.hInstance     := HInstance;
    wc.hIcon         := Icon;
    wc.hCursor       := Cursor;
    wc.hbrBackGround := Brush;
    wc.lpszClassName := PChar(AClassName);
    if RegisterClass(wc) = 0 then
      Exit;
    FRegisteredClass := True;
  end;

  FHandle := CreateWindowEx(ExStyle, PChar(AClassName), PChar(Caption), Style,
    ALeft, ATop, AWidth, AHeight, 0, 0, HInstance, Self);
end;

procedure TSForm.SetStayOnTop(Value: Boolean);
begin
  FStayOnTop := Value;
  if FHandle = 0 then
    Exit;
  if FStayOnTop then
    SetWindowPos(FHandle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE)
  else
    SetWindowPos(FHandle, HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE);
end;

procedure TSForm.SetFullScreen(Value: Boolean);
begin
  FFullScreen := Value;
  if FHandle = 0 then
    Exit;
  if FFullScreen then
  begin
    FSaveStayOnTop := StayOnTop;
    StayOnTop := True;
    FillChar(FSaveWP, SizeOf(FSaveWP), 0);
    FSaveWP.length := SizeOf(FSaveWP);
    GetWindowPlacement(FHandle, @FSaveWP);
    FSaveWS := GetWindowLong(FHandle, GWL_STYLE);
    SetWindowLong(FHandle, GWL_STYLE, WS_OVERLAPPEDWINDOW
      and not WS_CAPTION and not WS_THICKFRAME and not WS_MAXIMIZEBOX);
    ShowWindow(FHandle, SW_SHOWNA);
    SetBounds(0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN));
  end else
  begin
    SetWindowLong(FHandle, GWL_STYLE, FSaveWS and not WS_MAXIMIZE);
    ShowWindow(FHandle, SW_SHOWNA);
    SetWindowPlacement(FHandle, @FSaveWP);
    StayOnTop := FSaveStayOnTop;
  end;
end;

// StyleWS_OVERLAPPEDWINDOWȊÔƂ͐삵Ȃ
procedure TSForm.SetBorder(Value: Boolean);
var
  rc: TRect;
begin
  FBorder := Value;
  rc := ClientRect;
  if FBorder then
    SetWindowLong(FHandle, GWL_STYLE, WS_OVERLAPPEDWINDOW)
  else
    SetWindowLong(FHandle, GWL_STYLE, WS_OVERLAPPEDWINDOW
      and not WS_CAPTION and not WS_THICKFRAME and not WS_MAXIMIZEBOX);
  ShowWindow(FHandle, SW_SHOWNA);
  ClientRect := rc;
end;

function TSForm.GetMaximized: Boolean;
begin
  Result := IsZoomed(FHandle);
end;

function TSForm.GetMinimized: Boolean;
begin
  Result := IsIconic(FHandle);
end;

procedure TSForm.Show;
begin
  inherited;
  Update;
end;

procedure TSForm.Close;
begin
  Perform(WM_SYSCOMMAND, SC_CLOSE, 0);
end;

procedure TSForm.Restore;
begin
  ShowWindow(FHandle, SW_RESTORE);
end;

{ TSDlg }

function SDlgProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): BOOL; stdcall;
var
  Dlg: TSDlg;
  Message: TMessage;
begin
   Message.Msg := Msg;
   Message.WParam := wParam;
   Message.LParam := lParam;
   Message.Result := 0;
   case Msg of
     WM_INITDIALOG:
       begin
         SetProp(hWnd, PropName, lParam);
         TSDlg(lParam).Handle := hWnd;
         Message.Result := 1;
         TSDlg(lParam).Dispatch(Message);
         Result := BOOL(Message.Result);
       end;
     WM_NCDESTROY:
       begin
         TSDlg(GetProp(hWnd, PropName)).Handle := 0;
         RemoveProp(hWnd, PropName);
         Result := True;
       end;
   else
     Dlg := TSDlg(GetProp(hWnd, PropName));
     if Assigned(Dlg) then
       Dlg.Dispatch(Message);
     Result := BOOL(Message.Result);
   end;
end;

constructor TSDlg.Create;
begin
  FItem := TWinControl.Create;
end;

constructor TSDlg.CreateByResource(Parent: HWND; const TemplateName: string);
begin
  Create;
  FParent := Parent;
  FTemplateName := TemplateName;
end;

destructor TSDlg.Destroy;
begin
  FItem.Free;
  inherited;
end;

function TSDlg.GetItem(Id: Integer): TWinControl;
begin
  FItem.Handle := GetDlgItem(Handle, Id);
  Result := FItem;
end;


{ TSModalDlg }

function TSModalDlg.ShowModal: Boolean;
begin
  Result := DialogBoxParam(HInstance, PChar(FTemplateName), FParent, @SDlgProc,
    LParam(Self)) = IDOK;
end;

procedure TSModalDlg.Close(ResultCode: Integer);
begin
  EndDialog(FHandle, ResultCode);
end;


{ TSModelessDlg }

procedure TSModelessDlg.ShowModeless;
begin
  CreateDialogParam(hInstance, PChar(FTemplateName), FParent, @SDlgProc, LParam(Self));
end;

procedure TSModelessDlg.Close;
begin
  DestroyWindow(FHandle);
end;

{ TSPropSheet }

function SPropProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  Dlg: TSDlg;
  Message: TMessage;
  psp: PPropSheetPage;
begin
   Message.Msg := Msg;
   Message.WParam := wParam;
   Message.LParam := lParam;
   Message.Result := 0;
   case Msg of
     WM_INITDIALOG:
       begin
         psp := PPropSheetPage(lParam);
         SetProp(hWnd, PropName, psp.lParam);
         TSDlg(psp.lParam).Handle := hWnd;
         Message.Result := 1;
         TSDlg(psp.lParam).Dispatch(Message);
         Result := Message.Result;
       end;
     WM_NCDESTROY:
       begin
         TSDlg(GetProp(hWnd, PropName)).Handle := 0;
         RemoveProp(hWnd, PropName);
         Result := 1;
       end
   else
     Dlg := TSDlg(GetProp(hWnd, PropName));
     if Assigned(Dlg) then
       Dlg.Dispatch(Message);
     Result := Message.Result;
   end;
end;

constructor TSPropSheet.Create(Parent: HWND; const Caption: string; Pages: array of TSDlg);
var
  i: Integer;
  p: PPropSheetPage;
begin
  FCaption := Caption;
  GetMem(FPages, SizeOf(TPropSheetPage) * (High(Pages)+1));
  p := FPages;

  for i := 0 to High(Pages) do
  begin
    FillChar(p^, SizeOf(p^), 0);
    p.dwSize := SizeOf(p^);
    p.hInstance := HInstance;
    p.pszTemplate := PChar(Pages[i].FTemplateName);
    p.pfnDlgProc := @SPropProc;
    p.lParam := LParam(Pages[i]);
    Inc(p);
  end;

  FHeader.dwSize := SizeOf(FHeader);
  FHeader.dwFlags := PSH_PROPSHEETPAGE or PSH_NOAPPLYNOW;
  FHeader.hwndParent := Parent;
  FHeader.hInstance := HInstance;
  FHeader.pszCaption := PChar(FCaption);
  FHeader.nPages := High(Pages)+1;
  FHeader.ppsp := FPages;
end;

destructor TSPropSheet.Destroy;
begin
  FreeMem(FPages);
  inherited;
end;

function TSPropSheet.ShowModal: Boolean;
begin
  Result := PropertySheet(FHeader) <> -1;
end;


{ TSOpenDlg }

constructor TSOpenDlg.Create(Parent: HWND; const Filter: array of string; const FileName: string);
var
  i: Integer;
begin
  FParent := Parent;
  for i := 0 to High(Filter) do
    FFilter := FFilter + Filter[i] + #0;
  FFilter := FFilter + #0;
  FFileName := FileName;
end;

function TSOpenDlg.ShowModal(var AFileName: string): Boolean;
var
  ofn: TOpenFilename;
  buf: array[0..MAX_PATH] of Char;
begin
  FillChar(ofn, SizeOf(ofn), 0);
  FillChar(buf, SizeOf(buf), 0);
  with ofn do
  begin
    lStructSize := SizeOf(ofn);
    hWndOwner := FParent;
    lpstrFilter := PChar(FFilter);
    lpstrFile := buf;
    nMaxFile := SizeOf(buf);
    Flags := OFN_FILEMUSTEXIST or OFN_HIDEREADONLY;
  end;
  Move(PChar(FFileName)^, buf, Length(FFileName));
  Result :=  GetOpenFileName(ofn);
  AFileName := ofn.lpstrFile;
end;


{ TSSaveDlg }

function TSSaveDlg.ShowModal(var AFileName: string): Boolean;
var
  ofn: TOpenFilename;
  buf: array[0..MAX_PATH] of Char;
begin
  FillChar(ofn, SizeOf(ofn), 0);
  FillChar(buf, SizeOf(buf), 0);
  with ofn do
  begin
    lStructSize := SizeOf(ofn);
    hWndOwner := FParent;
    lpstrFilter := PChar(FFilter);
    lpstrFile := buf;
    nMaxFile := SizeOf(buf);
    Flags := OFN_OVERWRITEPROMPT or OFN_HIDEREADONLY;
  end;
  Move(PChar(FFileName)^, buf, Length(FFileName));
  Result :=  GetSaveFileName(ofn);
  AFileName := ofn.lpstrFile;
end;


{ TSMenu }

constructor TSMenu.Create;
begin
  FHandle := CreateMenu;
end;

destructor TSMenu.Destroy;
begin
  DestroyMenu(FHandle);
  inherited Destroy;
end;

function TSMenu.GetFlag(const Caption: string; Id: Integer): Integer;
begin
  if IsMenu(Id) then
    Result := MF_POPUP
  else
    Result := MF_STRING;
  if Caption = '-' then
    Result := Result or MF_SEPARATOR;
end;

procedure TSMenu.AppendItem(const Caption: string; Id: Integer);
begin
  AppendMenu(FHandle, GetFlag(Caption, Id), Id, PChar(Caption));
end;

procedure TSMenu.InsertItem(Index: Integer; const Caption: string; Id: Integer);
begin
  InsertMenu(FHandle, Index, GetFlag(Caption, Id) or MF_BYPOSITION, Id, PChar(Caption));
end;

procedure TSMenu.GrayItem(Id: Integer);
begin
  EnableMenuItem(FHandle, Id, MF_GRAYED);
end;

procedure TSMenu.GrayItemByPos(Index: Integer);
begin
  EnableMenuItem(FHandle, Index, MF_GRAYED or MF_BYPOSITION);
end;

procedure TSMenu.CheckItem(Id: Integer; Checked: Boolean);
begin
  if Checked then
    CheckMenuItem(FHandle, Id, MF_CHECKED)
  else
    CheckMenuItem(FHandle, Id, MF_UNCHECKED);
end;

function TSMenu.IndexOfItem(Id: Integer): Integer;
var
  i, cnt: Integer;
begin
  cnt := GetMenuItemCount(FHandle);
  for i := 0 to cnt - 1 do
  begin
    if GetMenuItemID(FHandle, i) = UINT(Id) then
    begin
      Result := i;
      Exit;
    end;
  end;
  Result := -1;
end;

{ TSPopupMenu }

constructor TSPopupMenu.Create;
begin
  FHandle := CreatePopupMenu;
end;

procedure TSPopupMenu.Popup(Parent: HWND; const pt: TPoint);
begin
  TrackPopupMenu(FHandle, TPM_LEFTALIGN or TPM_LEFTBUTTON or TPM_RIGHTBUTTON,
    pt.x, pt.y, 0, Parent, nil);
end;

function GetContextMenuPos(Handle: HWND; LParam: LPARAM; SelPosX, SelPosY: Longint; var pt: TPoint): Boolean;
var
  rc: TRect;
begin
  if LParam = Longint($FFFFFFFF) then
  begin
    pt.x := SelPosX;
    pt.y := SelPosY;
  end else
  begin
    pt.x := LoWord(LParam);
    pt.y := HiWord(LParam);
    ScreenToClient(Handle, pt);
  end;
  GetClientRect(Handle, rc);
  Result := PtInRect(rc, pt);
  ClientToScreen(Handle, pt);
end;


{ TSTooltip }

constructor TSTooltip.Create(Parent: HWnd);
var
  ti: TToolInfo;
begin
  FParent := Parent;
  FHandle:= CreateWindowEx(0, 'tooltips_class32', nil, TTS_ALWAYSTIP,
    Integer(CW_USEDEFAULT), 0, Integer(CW_USEDEFAULT), 0, FParent, 0, HInstance, nil);
  if FHandle = 0 then
    Exit;
  FillChar(ti, SizeOf(ti), 0);
  with ti do
  begin
    cbSize := SizeOf(ti);
    uFlags := TTF_SUBCLASS;
    hwnd   := FParent;
    hinst  := HInstance;
    rect.Right := 64;
    rect.Bottom := 64;
  end;
  SendMessage(FHandle, TTM_ADDTOOL, 0, Integer(@ti));
end;

destructor TSTooltip.Destroy;
begin
  if FHandle <> 0 then
    DestroyWindow(FHandle);
end;

procedure TSTooltip.Show;
begin
  SendMessage(FHandle, TTM_ACTIVATE, WPARAM(True) ,0);
end;

procedure TSTooltip.Hide;
begin
  SendMessage(FHandle, TTM_ACTIVATE, WPARAM(False) ,0);
end;

procedure TSTooltip.SetTime(const Value: Integer);
begin
  SendMessage(FHandle, TTM_SETDELAYTIME, TTDT_INITIAL, Value);
  //SendMessage(FHandle, TTM_SETDELAYTIME, TTDT_AUTOMATIC, Value);
end;

procedure TSTooltip.SetText(const Value: string);
var
  ti: TToolInfo;
begin
  with ti do
  begin
    cbSize   := SizeOf(ti);
    uId      := 0;
    lpszText := PChar(Value);
    hinst    := HInstance;
    hwnd     := FParent;
    SendMessage(FHandle, TTM_UPDATETIPTEXT, 0, Integer(@ti));
  end;
end;

procedure TSTooltip.UpdateRect;
var
  ti: TToolInfo;
begin
  with ti do
  begin
    cbSize := SizeOf(ti);
    hwnd   := FParent;
    uId    := 0;
    GetClientRect(FParent, rect);
  end;
  SendMessage(FHandle, TTM_NEWTOOLRECT, 0, Integer(@ti));
end;

procedure TSTooltip.SetRect(const Rect: TRect);
var
  ti: TToolInfo;
begin
  ti.cbSize := SizeOf(ti);
  ti.hwnd := FParent;
  ti.uId := 0;
  ti.Rect := Rect;
  SendMessage(FHandle, TTM_NEWTOOLRECT, 0, Integer(@ti));
end;


{ Function and Procedure }

function MessageLoop: Integer;
var
  Msg: TMsg;
  Ret: Integer;
begin
  while True do
  begin
    Ret := Integer(GetMessage(Msg, 0, 0, 0));
    if (Ret = 0) or (Ret = -1) then
      Break
    else begin
      if (MsgMouseWheel <> 0) and (Msg.message = MsgMouseWheel) then
        Msg.message := WM_MOUSEWHEEL;
      TranslateMessage(Msg);
      DispatchMessage(Msg);
    end;
  end;
  Result := Msg.wParam;
end;

function MessageLoopWithAccel(hWnd: HWND; var hAccel: HACCEL): Integer;
var
  Msg: TMsg;
  Ret: Integer;
begin
  while True do
  begin
    Ret := Integer(GetMessage(Msg, 0, 0, 0));
    if (Ret = 0) or (Ret = -1) then
      Break
    else if not LongBool(TranslateAccelerator(hWnd, hAccel, Msg)) then
    begin
      if (MsgMouseWheel <> 0) and (Msg.message = MsgMouseWheel) then
        Msg.message := WM_MOUSEWHEEL;
      TranslateMessage(Msg);
      DispatchMessage(Msg);
    end;
  end;
  Result := Msg.wParam;
end;

function MessageLoopWithDialog(var hDlg: HWND): Integer;
var
  Msg: TMsg;
  Ret: Integer;
begin
  while True do
  begin
    Ret := Integer(GetMessage(Msg, 0, 0, 0));
    if (Ret = 0) or (Ret = -1) then
      Break
    else if (hDlg = 0) or not IsDialogMessage(hDlg, Msg) then
    begin
      if (MsgMouseWheel <> 0) and (Msg.message = MsgMouseWheel) then
        Msg.message := WM_MOUSEWHEEL;
      TranslateMessage(Msg);
      DispatchMessage(Msg);
    end;
  end;
  Result := Msg.wParam;
end;

function MessageLoopWithAccelAndDialog(hWnd: HWND; hAccel: HACCEL; var hDlg: HWND): Integer;
var
  Msg: TMsg;
  Ret: Integer;
begin
  while True do
  begin
    Ret := Integer(GetMessage(Msg, 0, 0, 0));
    if (Ret = 0) or (Ret = -1) then
      Break
    else if (hDlg = 0) or not IsDialogMessage(hDlg, Msg) then
    begin
      if not LongBool(TranslateAccelerator(hWnd, hAccel, Msg)) then
      begin
        if (MsgMouseWheel <> 0) and (Msg.message = MsgMouseWheel) then
          Msg.message := WM_MOUSEWHEEL;
        TranslateMessage(Msg);
        DispatchMessage(Msg);
      end;
    end;
  end;
  Result := Msg.wParam;
end;

function AlreadyRun(const AppName: string): Boolean;
begin
  if CreateMutex(nil, True, PChar(AppName+'Mutex')) <> 0 then
    Result := GetLastError = ERROR_ALREADY_EXISTS
  else begin
    Result := False;
    MessageBox(GetDesktopWindow, 'CreateMutex failed.', PChar(AppName), MB_OK or MB_ICONSTOP);
    Halt(255);
  end;
end;

procedure FormatMsgBox(hWnd: HWND; const Caption: string; const Text: string;
  const Args: array of string; Flags: Integer);
type
  TPointerArray = array[0..0] of Pointer;
  PPointerArray = ^TPointerArray;
var
  pBuf: PChar;
  i: Integer;
  TmpArg: PPointerArray;
begin
  if not IsWindowVisible(hWnd) then
    hWnd := GetDesktopWindow;

  GetMem(TmpArg, SizeOf(Pointer) * (High(Args)+1));
  try
    for i := 0 to High(Args) do
      TmpArg[i] := PChar(Args[i]);
    FormatMessage(FORMAT_MESSAGE_FROM_STRING or FORMAT_MESSAGE_ARGUMENT_ARRAY or
      FORMAT_MESSAGE_ALLOCATE_BUFFER, PChar(Text), 0, 0, PChar(@pBuf), 0, TmpArg);
  finally
    FreeMem(TmpArg);
  end;
  MessageBox(hWnd, pBuf, PChar(Caption), Flags);
  LocalFree(HLOCAL(pBuf));
end;

procedure SetWindowCenter(hWnd: HWND);
var
  rc, drc: TRect;
begin
  GetWindowRect(GetDesktopWindow, drc);
  GetWindowRect(hWnd, rc);
  SetWindowPos(hWnd, 0, (drc.Right - (rc.Right - rc.Left)) div 2,
    (drc.Bottom - (rc.Bottom - rc.Top)) div 2, 0, 0,
    SWP_NOZORDER or SWP_NOSIZE or SWP_NOACTIVATE);
end;

function OsVersionInfo: TOsVersionInfo;
begin
  FillChar(Result, SizeOf(Result), 0);
  Result.dwOSVersionInfoSize := SizeOf(TOsVersionInfo);
  GetVersionEx(Result);
end;

procedure AboutBox(hWnd: HWND; const Caption: string; const Text: string);
const
  ABIconW: PWideChar = 'MAINICON';
  ABIconA: PChar = 'MAINICON';
var
  mbp: TMsgBoxParams;
begin
  FillChar(mbp, SizeOf(mbp), 0);
  with mbp do
  begin
    cbSize := SizeOf(mbp);
    hwndOwner := hWnd;
    hInstance := MainInstance;
    lpszText := PChar(Text);
    lpszCaption := PChar(Caption);
    dwStyle := MB_USERICON;
    // NTłWideCharœnȂƂ(ANSI version APIȂ̂Ɂc)
    if OsVersionInfo.dwPlatformId = VER_PLATFORM_WIN32_NT then
       lpszIcon := Pointer(ABIconW)
    else
       lpszIcon := ABIconA;
  end;
  MessageBoxIndirect(mbp);
end;

function MsgBox(hWnd: HWND; const Caption: string; const Text: string): Integer;
begin
  Result := MessageBox(hWnd, PChar(Text), PChar(Caption), MB_OK);
end;

procedure ShowMsg(const Msg: string);
begin
  MessageBox(0, PChar(Msg), PChar(ParamStr(0)), MB_OK);
end;

function Rect(ALeft, ATop, ARight, ABottom: Integer): TRect;
begin
  with Result do
  begin
    Left := ALeft;
    Top := ATop;
    Right := ARight;
    Bottom := ABottom;
  end;
end;

function Point(AX, AY: Integer): TPoint;
begin
  with Result do
  begin
    X := AX;
    Y := AY;
  end;
end;

begin
  MsgMouseWheel := RegisterWindowMessage('MSWHEEL_ROLLMSG'); // for Win95
end.
